//-----------------------------------------------------------------------------
// File:     StandingLegForVisualization.cpp
// Class:    None
// Parent:   None
// Children: None
// Purpose:  Simulates a 2D human composed of 4 rigid bodies (foot, shank, thigh, HAT)
// Authors:  Edith Arnold & Melanie Fox
//-----------------------------------------------------------------------------
// The following are standard C/C++ header files.
// If a filename is enclosed inside < >  it means the header file is in the Include directory.
// If a filename is enclosed inside " "  it means the header file is in the current directory.
#include <ctype.h>      // Character Types
#include <math.h>       // Mathematical Constants
#include <stdarg.h>     // Variable Argument Lists
#include <stdio.h>      // Standard Input/Output Functions
#include <stdlib.h>     // Utility Functions
#include <string.h>     // String Operations
#include <signal.h>     // Signals (Contol-C + Unix System Calls)
#include <setjmp.h>     // Nonlocal Goto (For Control-C)
#include <time.h>       // Time and Date information
#include <assert.h>     // Verify Program Assertion
#include <errno.h>      // Error Codes (Used in Unix system())
#include <float.h>      // Floating Point Constants
#include <limits.h>     // Implementation Constants
#include <stddef.h>     // Standard Definitions
#include <exception>    // Exception handling (e.g., try, catch throw)
//-----------------------------------------------------------------------------
#include "SimTKsimbody.h"
#include "ScaledBodySegmentMassProperties.h"
#include "UserForceControlJointTorquesFinal.h"
using namespace SimTK;
using namespace std;
//-----------------------------------------------------------------------------
#define PI 3.14159265358979

//-----------------------------------------------------------------------------
// Prototypes for local functions (functions not called by code in other files)
//-----------------------------------------------------------------------------
bool  Simulate2DHuman( void );
bool  WriteStringToFile(   const char outputString[], FILE *fptr )  { return fputs( outputString, fptr ) != EOF; }
bool  WriteStringToScreen( const char outputString[] )              { return WriteStringToFile( outputString, stdout ); }
bool  WriteStringToScreenAndFile( const char s[], FILE *fptr )      { return WriteStringToScreen( s )  &&  WriteStringToFile( s, fptr ); }
bool  WriteDoubleToFile( double x, int precision, FILE *fptr );
bool  WriteDoubleToScreenAndFile( double x, int precision, FILE *fptr )  { return WriteDoubleToFile( x, precision, stdout ) && WriteDoubleToFile( x, precision, fptr ); }
bool  GetScaledSegmentMassAndInertiaProperties( bool subjectIsFemale, const char *segmentName, Real subjectTotalMassInKG, Real subjectTotalHeightInMM, Real subjectSegmentLengthInMM, MassProperties &segmentMassProperties, Real &segmentScaledLengthInM );
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute );
double ConvertFromRadiansToDegrees( double angleInRadians ) {return( angleInRadians*180.0/PI );}
double ConvertFromDegreesToRadians( double angleInDegrees ) {return( angleInDegrees*PI/180.0 );}


//-----------------------------------------------------------------------------
// The executable program starts here
//-----------------------------------------------------------------------------
int  main( int numberOfCommandLineArguments, char *arrayOfCommandLineArguments[] )
{
   // Default value is program failed
   bool programSucceeded = false;

   // It is a good programming practice to do little in the main function of a program.
   // The try-catch code in this main routine catches exceptions thrown by functions in the
   // try block, e.g., catching an exception that occurs when a NULL pointer is de-referenced.
   try
   {
      // Do the required programming tasks
      programSucceeded = Simulate2DHuman();

   }
   // This catch statement handles certain types of exceptions
   catch( const exception &e )
   {
      WriteStringToScreen( "\n\n Error: Programming error encountered.\n The exception thrown is: " );
      WriteStringToScreen( e.what() );
   }
   // The exception-declaration statement (...) handles any type of exception,
   // including C exceptions and system/application generated exceptions.
   // This includes exceptions such as memory protection and floating-point violations.
   // An ellipsis catch handler must be the last handler for its try block.
   catch( ... )
   {
      WriteStringToScreen( "\n\n Error: Programming error encountered.\n An unhandled exception was thrown." );
   }

   // Give the user a chance to see a final message before exiting the program.
   WriteStringToScreen( programSucceeded ? "\n\n Program succeeded.  Press  Enter  to finish: " : "\n\n Program failed.  Press  Enter  to finish: " );
   getchar();  // Keeps the screen displayed until the user presses the Enter (or Return) key.

   // The value returned by the main function is the exit status of the program.
   // A normal program exit returns 0 (other return values usually signal an error).
   return programSucceeded == true ? 0 : 1;
}


//-----------------------------------------------------------------------------
bool  Simulate2DHuman( void )
{
   // Declare a multibody system
   // Each multibody system contains at most one SimbodyMatterSubsytem (which inherits from MatterSubsystem - the ma in F=ma)
   // Each multibody system can contain 0, 1, 2, ... force subsystems (the F in F=ma)
   MultibodySystem  mbs;
   SimbodyMatterSubsystem  human;      // Create a single matter sub-system (for the apple)

   // 0. The ground's right-handed, orthogonal x,y,z unit vectors are directed with x horizontally right and y vertically upward.
   // 1. Create a gravity vector that is straight down (in the ground's frame)
   // 2. Create a uniform gravity sub-system
   // 3. Add the gravity sub-system to the multibody system
   Vec3 gravityVector( 0, -9.8, 0 );
   UniformGravitySubsystem gravity( gravityVector );
   mbs.addForceSubsystem( gravity );

   // Create the mass, center of mass, and inertia properties for the book
   const Real subjectTotalMassInKG = 56.7;
   const Real subjectTotalHeightInMM = 1651;
   const bool subjectIsFemale = true;

   const char *segmentName[] =      { "Head", "Trunk",  "UpperArm",  "ForeArm",   "Hand",  "Thigh",   "Shank",  "Foot" };
   const Real measuredSegmentLengthInMM[] = {    0.0,     0.0,         0.0,        0.0,      0.0,      0.0,       0.0,    0.0 };

   MassProperties footMassProperties;
   MassProperties shankMassProperties;
   MassProperties thighMassProperties;
   MassProperties handMassProperties;
   MassProperties foreArmMassProperties;
   MassProperties upperArmMassProperties;
   MassProperties trunkMassProperties;
   MassProperties headMassProperties;

   Real footLength;
   Real shankLength;
   Real thighLength;
   Real handLength;
   Real foreArmLength;
   Real upperArmLength;
   Real trunkLength;
   Real headLength;

   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[7], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[7], footMassProperties,     footLength );
   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[6], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[6], shankMassProperties,    shankLength );
   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[5], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[5], thighMassProperties,    thighLength );
   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[4], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[4], handMassProperties,     handLength );
   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[3], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[3], foreArmMassProperties,  foreArmLength );
   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[2], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[2], upperArmMassProperties, upperArmLength );
   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[1], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[1], trunkMassProperties,    trunkLength );
   GetScaledSegmentMassAndInertiaProperties( subjectIsFemale, segmentName[0], subjectTotalMassInKG, subjectTotalHeightInMM, measuredSegmentLengthInMM[0], headMassProperties,     headLength );

   // Lump head, arms, and torso together as HAT.  Approximate COM as torso COM.
   Real HATMass = headMassProperties.getMass() + trunkMassProperties.getMass() + upperArmMassProperties.getMass() + foreArmMassProperties.getMass() + handMassProperties.getMass();
   Real trunkMass = trunkMassProperties.getMass();
   Real massScalingFactorHAT = 0.5 * HATMass / trunkMass;
   const Inertia HATInertiaAboutOrigin = trunkMassProperties.getInertia() * massScalingFactorHAT;
   const Vec3 HATCenterOfMassLocation = trunkMassProperties.getMassCenter();
   Real HATLength = trunkLength;
   //HATMass = 23.0;
   const MassProperties  HATMassProperties( HATMass, HATCenterOfMassLocation, HATInertiaAboutOrigin );

   // The motion of the segment is related to the motion of the ground via "mobilizers"
   // The "mobilizers" specify the allowable motion of the segment to the ground.
   // More specifically, the motion of an "outboard body" (e.g., the segment)
   // to its inboard body (e.g., the ground) is specified by first constructing:
   // 1. An  "outboard frame"  on the ground (which hooks to the "outboard body" - the segment )
   // 2. An  "inboard  frame"  on the segment  (which hooks to the "inboard  body" - the ground)

   // The orientation and position of the  outboard frame  from the ground's frame is specified below.
   // The outboard frame's axes are aligned with the ground's axes and its origin is coincident with the ground's origin.
   // In other words, for this simple problem the outboard frame and the ground frame are identical.

   // Outboard frame for Shank to Ground Mobilizer (Ankle)
   const Rotation   outboardFrameOrientationInShank;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       outboardFrameOriginLocationFromShankOrigin( 0, 0, 0 );
   const Transform  outboardFrameTransformFromShank( outboardFrameOrientationInShank, outboardFrameOriginLocationFromShankOrigin );
   
   // Outboard frame for Thigh to Shank Mobilizer (Knee)
   const Rotation   outboardFrameOrientationInThigh;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       outboardFrameOriginLocationFromThighOrigin( 0, 0, 0 );
   const Transform  outboardFrameTransformFromThigh( outboardFrameOrientationInThigh, outboardFrameOriginLocationFromThighOrigin );

   // Outboard frame for HAT to Thigh Mobilizer (Hip)
   const Rotation   outboardFrameOrientationInHAT;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       outboardFrameOriginLocationFromHATOrigin( 0, 0, 0 );
   const Transform  outboardFrameTransformFromHAT( outboardFrameOrientationInHAT, outboardFrameOriginLocationFromHATOrigin );


   // The orientation and position of the  inboard  frame  from the segment's  frame is specified below.
   // The inboard frame's  axes are aligned with the segment's  axes and its origin is coincident with the segment's  origin
   // In other words, for this simple problem the inboard  frame and the segment frame are identical.
   // Although the inboard frame can be constructed in a simple manner analogous to the outboardFrameTransformFromGround (above)
   // it is worthwhile to look at the details of the rotation matrix and position vector in the transform:
   // a. The rotation matrix relating the InboardFrame's x,y,z axes to the SegmentFrame's x,y,z axes is specified InboardFrame_SegmentFrame.
   // b. The position of the InboardFrame's origin from the SegmentFrame origin, expressed in terms of the SegmentFrame's "x, y, z" unit vectors.
  
   // Inboard frame for Ground to Shank Mobilizer
   const Rotation   inboardFrameOrientationInGround;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromGroundOrigin( 0, 0, 0 );
   const Transform  inboardFrameTransformFromGround( inboardFrameOrientationInGround, inboardFrameOriginLocationFromGroundOrigin );
   
   // Inboard frame for Shank to Thigh Mobilizer
   const Rotation   inboardFrameOrientationInShank;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromShankOrigin( 0, shankLength, 0 );
   const Transform  inboardFrameTransformFromShank( inboardFrameOrientationInShank, inboardFrameOriginLocationFromShankOrigin );

   // Inboard frame for Thigh to HAT Mobilizer
   const Rotation   inboardFrameOrientationInThigh;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromThighOrigin( 0, thighLength, 0 );
   const Transform  inboardFrameTransformFromThigh( inboardFrameOrientationInThigh, inboardFrameOriginLocationFromThighOrigin );
  
   // The model is a 2D simplification, with the rigid bodies moving in the saggital plane so we specify planar mobilizers

   //Mobilizer footToGroundMobilizer = Mobilizer::Pin;
   //const BodyId footBodyId = human.addRigidBody( footMassProperties, inboardFrameTransformFromFoot, GroundId, outboardFrameTransformFromGround, footToGroundMobilizer );

   Mobilizer shankToGroundMobilizer = Mobilizer::Pin;
   const BodyId shankBodyId = human.addRigidBody( shankMassProperties, outboardFrameTransformFromShank, GroundId, inboardFrameTransformFromGround, shankToGroundMobilizer );

   Mobilizer thighToShankMobilizer = Mobilizer::Pin;
   const BodyId thighBodyId = human.addRigidBody( thighMassProperties, outboardFrameTransformFromThigh, shankBodyId, inboardFrameTransformFromShank, thighToShankMobilizer );

   Mobilizer HATToThighMobilizer = Mobilizer::Pin;
   const BodyId HATBodyId = human.addRigidBody( HATMassProperties, outboardFrameTransformFromHAT, thighBodyId, inboardFrameTransformFromThigh, HATToThighMobilizer );

   // Add the matter sub-system to the system.
   mbs.setMatterSubsystem( human );    

   // Add user force elements to the multibody system
   GeneralForceElements userForceElements;
   mbs.addForceSubsystem( userForceElements );

   // Although "new" was used to allocate this UserForce, do not "delete it.
   // this bug will be fixed in the next version of Simbody so it can take an object from the stack or heap.
   // For now, addUSerForce takes ownership of the allocated item and takes care of deleting it at the end.


   //FILE *outputFile = FileOpenWithMessageIfCannotOpen( "MipsiControlTorques.txt", "w" );
   // Computed torque control Force
   UserForceControlJointTorques *computedTorque = new UserForceControlJointTorques( shankBodyId, thighBodyId, HATBodyId, 1.0, 60.0 );
   userForceElements.addUserForce( computedTorque );


   //Create a state for this system.
   //Define appropriate states for this multi-body system.
   //Set the initial time to 0.0
   State s;
   mbs.realize( s );
   s.setTime( 0.0 );

   // Set the initial values for the configuration variables (ankle, knee, & hip angles and velocities)
   human.setMobilizerQ( s, shankBodyId , 0,  ConvertFromDegreesToRadians( -10.0 ) );
   human.setMobilizerU( s, shankBodyId , 0,  0.0 );

   human.setMobilizerQ( s, thighBodyId , 0,  ConvertFromDegreesToRadians( 10.0 ) );
   human.setMobilizerU( s, thighBodyId , 0,  0.0 );

   human.setMobilizerQ( s, HATBodyId , 0,  ConvertFromDegreesToRadians( -10.0 ) );
   human.setMobilizerU( s, HATBodyId , 0,  0.0 );

   // Create a study using the Runge Kutta Merson integrator (alternately use the CPodesIntegrator)
   RungeKuttaMerson myStudy( mbs, s );

   // Set the numerical accuracy for the integrator
   myStudy.setAccuracy( 1.0E-7 );

   // The next statement does lots of accounting
   myStudy.initialize();

   // Open a file to record the simulation results (they are also displayed on screen)
   //FILE *outputFile = FileOpenWithMessageIfCannotOpen( "MipsiResults5.txt", "w" );
   //WriteStringToScreenAndFile(   "Time      Ankle Angle      Knee Angle      HipAngle \n",  outputFile );

   // Visualize results with VTKReporter
   VTKReporter animationResults( mbs );

   //  Define cylinder radii for visualization of each segment
   Real shankRadius = shankLength/8.0;
   Real thighRadius = thighLength/8.0;
   Real HATRadius = HATLength/3.0;

   DecorativeCylinder shankCylinder = DecorativeCylinder( shankRadius, shankLength*0.5 );
   shankCylinder.setColor(Blue);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   shankCylinder.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   DecorativeCylinder thighCylinder = DecorativeCylinder( thighRadius, thighLength*0.5 );
   thighCylinder.setColor(Red);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   thighCylinder.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   DecorativeCylinder HATCylinder = DecorativeCylinder( HATRadius, HATLength*0.5 );
   HATCylinder.setColor(Green);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   HATCylinder.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   // Decorate each segment with a cylinder at the segment's origin
   const Rotation  segmentToCylinderOrientation; //( 1,0,0, 0,1,0, 0,0,1 );
   
   const Vec3      shankOriginToShankCylinderOriginLocation( 0, shankLength*0.5, 0 );
   const Transform segmentToShankCylinderTransform( segmentToCylinderOrientation, shankOriginToShankCylinderOriginLocation );

   const Vec3      thighOriginToThighCylinderOriginLocation( 0, thighLength*0.5, 0 );
   const Transform segmentToThighCylinderTransform( segmentToCylinderOrientation, thighOriginToThighCylinderOriginLocation );
   
   const Vec3      HATOriginToHATCylinderOriginLocation( 0, HATLength*0.5, 0 );
   const Transform segmentToHATCylinderTransform( segmentToCylinderOrientation, HATOriginToHATCylinderOriginLocation );


   // animationResults.addDecoration( footBodyId,  segmentToRedBrickTransform, footBrick );
   animationResults.addDecoration( shankBodyId, segmentToShankCylinderTransform, shankCylinder );
   animationResults.addDecoration( thighBodyId, segmentToThighCylinderTransform, thighCylinder );
   animationResults.addDecoration( HATBodyId,   segmentToHATCylinderTransform,   HATCylinder );

   // Set the numerical integration step and the time for the simulation to run
   const Real dt = 0.01;
   const Real finalTime = 20.0;

   // Run the simulation and print the results
   while( 1 )
   {
      // Query for results to be printed
      Real time = s.getTime();
      //Real kineticEnergy = mbs.getKineticEnergy(s);
      //Real uniformGravitationalPotentialEnergy = mbs.getPotentialEnergy(s);
      //Real mechanicalEnergy = kineticEnergy + uniformGravitationalPotentialEnergy;
      

      // Query for each joint angle in radians.
      Real ankleAngle = human.getMobilizerQ( s, shankBodyId, 0 );
      Real kneeAngle = human.getMobilizerQ( s, thighBodyId, 0 );
      Real hipAngle = human.getMobilizerQ( s, HATBodyId, 0 );

      //// Print results to screen and file
      //WriteDoubleToScreenAndFile( time,                                     2, outputFile );
      //WriteDoubleToScreenAndFile( ConvertFromRadiansToDegrees(ankleAngle),  4, outputFile );
      //WriteDoubleToScreenAndFile( ConvertFromRadiansToDegrees(kneeAngle),   4, outputFile );
      //WriteDoubleToScreenAndFile( ConvertFromRadiansToDegrees(hipAngle),    4, outputFile );
      //WriteStringToScreenAndFile( "\n", outputFile );

      // Animate the results for this step
      animationResults.report(s);

      // Check if integration has completed
      if( time >= finalTime ) break;

      // Increment time step
      myStudy.step( time + dt);
   }

   // Simulation completed properly
   return true;
}


//-----------------------------------------------------------------------------
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute )
{
   // Try to open the file
   FILE *Fptr1 = fopen( filename, attribute );

   // If unable to open the file, issue a message
   if( !Fptr1 )
   {
      WriteStringToScreen( "\n\n Unable to open the file: " );
      WriteStringToScreen( filename );
      WriteStringToScreen( "\n\n" );
   }

   return Fptr1;
}


//-----------------------------------------------------------------------------
bool  WriteDoubleToFile( double x, int precision, FILE *fptr )
{
   // Ensure the precision (number of digits in the mantissa after the decimal point) makes sense.
   // Next, calculate the field width so it includes one extra space to the right of the number.
   if( precision < 0 || precision > 17 ) precision = 5;
   int fieldWidth = precision + 8;

   // Create the format specifier and print the number
   char format[20];
   sprintf( format, " %%- %d.%dE", fieldWidth, precision );
   return fprintf( fptr, format, x ) >= 0;
}
//-----------------------------------------------------------------------------
bool  GetScaledSegmentMassAndInertiaProperties( bool subjectIsFemale, const char *segmentName, Real subjectTotalMassInKG, Real subjectTotalHeightInMM, Real subjectSegmentLengthInMM, MassProperties &segmentMassProperties, Real &segmentScaledLengthInM )
{
   // Get the standard body segment mass properties
   const StandardBodySegmentMassProperties*  standardSegment = StandardBodySegmentMassProperties::getStandardHumanBodySegmentMassProperties( subjectIsFemale, segmentName );
   if( !standardSegment ) return false;

   // Scale the properties with the given information
   const ScaledBodySegmentMassProperties scaledSegment( *standardSegment, subjectTotalMassInKG, subjectTotalHeightInMM, subjectSegmentLengthInMM );

   // Store segment length and segment mass
   segmentScaledLengthInM = scaledSegment.getSegmentLengthInMM()/1000.0;

   Real segmentMassInKg = scaledSegment.getMassInKG();

   // Get the segment's distance from its origin anatomical marker to its mass center distance
   Real segmentMassCenterLength = scaledSegment.getMassCenterLengthInMM()/1000.0;

   // Get the segment x-radius of gyration about its mass center to the file
   // Calculate the segment x-moment of inertia  about its mass center to the file
   Real kxx = scaledSegment.getXRadiusOfGyrationInMM()/1000.0;
   Real segmentIxx = segmentMassInKg * pow(kxx,2);

   // Get the segment y-radius of gyration about its mass center to the file
   // Calculate the segment y-moment of inertia  about its mass center to the file
   Real kyy = scaledSegment.getYRadiusOfGyrationInMM()/1000.0;
   Real segmentIyy = segmentMassInKg * pow(kyy,2);

   // Get the segment z-radius of gyration about its mass center to the file
   // Calculate the segment z-moment of inertia  about its mass center to the file
   Real kzz = scaledSegment.getZRadiusOfGyrationInMM()/1000.0;
   Real segmentIzz = segmentMassInKg * pow(kzz,2);

   // The location of the segments' center of mass is a vector from the book's
   // origin expressed in the "x, y, z"' unit vectors fixed in the book's frame.
   // Example: The vector(0,0,0) locates the book's center of mass at the book's origin.
   // Example: The vector(1,0,0) locates the book's center of mass 1 unit in the "x" direction from the book's origin.
   const Vec3  segmentCenterOfMassLocation( 0, segmentScaledLengthInM-segmentMassCenterLength, 0);

   // Create the segments' inertia matrix about its origin for the "x, y, z" unit vectors fixed in the book's frame.
   // Note: The 3x3 inertia matrix is symmetric - so only 6 elements need to be defined.
   // Ixx, Iyy, Izz  are moments  of inertia (    diagonal terms in the matrix)
   // Ixy, Ixz, Iyz  are products of inertia (off-diagonal terms in the matrix)
   const Inertia  segmentInertiaAboutCM( segmentIxx, segmentIyy, segmentIzz );

   // Shift segment inertia matrix from CM to segment origin.  This format of inertia matrix is needed for MassProperties of segment.
   const Inertia  segmentInertiaAboutOrigin = segmentInertiaAboutCM.shiftFromMassCenter( -1*segmentCenterOfMassLocation, segmentMassInKg );

   // The MassProperties class holds the mass, center of mass, and inertia properties of a rigid body.
   // Although the next line creates an instance of the MassProperties class for each segment,
   // it does not get associated with the segment until the  addRigidBody  method.
   segmentMassProperties.setMassProperties( segmentMassInKg, segmentCenterOfMassLocation,  segmentInertiaAboutOrigin );

   return 0;
}


