//-----------------------------------------------------------------------------
// File:     UserForceControlJointTorques.h
// Class:    UserForceControlHorizontal
// Parent:   GeneralForceElements
// Children: None
// Purpose:  Applies joint torques to a 3 segment planar human model.
// Author:   Edith Arnold & Melanie Fox
//-----------------------------------------------------------------------------
#ifndef __USERFORCEVCONTROLHORIZONTAL_H__
#define __USERFORCEVCONTROLHORIZONTAL_H__
//-----------------------------------------------------------------------------
#include "StandardCppHeadersAndNamespace.h"
#include "SimTKsimbody.h"
using namespace SimTK;

//--------------------------------------------------------------------------
// User-defined classes for adding forces/torques are constructed as follows:
// 1. Create a constructor with whatever arguments make sense for the force or torque and copy the arguments into class data.
// 2. Create a clone method (all clone methods are identical except for the class name appearing after "new")
// 3. Create a calc  method (the arguments and return type for all calc methods are identical).
//    The code in the calc method is specific to the calculation of force or torque.
// Note: The set of all forces is replaced by an equivalent set, consisting of a torque
//       that is equal to the moment of the forces about the body's origin together
//       with the resultant of the forces applied at the body's origin.
//--------------------------------------------------------------------------
class UserForceControlJointTorques : public GeneralForceElements::UserForce 
{  
public:
   // Constructor is explicit
   explicit UserForceControlJointTorques( BodyId bodyIdA , BodyId bodyIdB, BodyId bodyIdC, Real damping, Real frequency)
   { 
      myShankId = bodyIdA;
      myThighId = bodyIdB;
      myHATId = bodyIdC;
      zeta = damping;
      omega = frequency;
   } 

   // The clone method is used internally by Simbody (required by virtual parent class)
   UserForce*  clone() const  { return new UserForceControlJointTorques(*this); }

   // The calc method is where forces or torques are calculated (required by virtual parent class)
   void  calc( const MatterSubsystem& matter,              // Input information (matter)
               const State&           state,               // Input information (current state)
               Vector_<SpatialVec>&   bodyForces,          // Forces and torques on bodies
               Vector_<Vec3>&         particleForces,      // Forces on particles (currently unused)
               Vector&                mobilityForces,      // Generalized forces 
               Real&                  pe ) const           // For forces with a potential energy
   {  
      // Query the matter subsystem for the joint angles and velocities.
      Real qA  = matter.getMobilizerQ( state, myShankId, 0 );
      Real qAp = matter.getMobilizerU( state, myShankId, 0 );		
      
      Real qB  = matter.getMobilizerQ( state, myThighId, 0 );
      Real qBp = matter.getMobilizerU( state, myThighId, 0 );		

      Real qC  = matter.getMobilizerQ( state, myHATId, 0 );
      Real qCp = matter.getMobilizerU( state, myHATId, 0 );		

      // Define constants (lengths & mass properties ) needed for torque calculation.
      Real AL    =  0.41137020172910665;    
      Real ALcm  =  0.22970912064553317;    
      Real BL    =  0.35065907780979827;    
      Real BLcm  =  0.22400101890489915;    
      Real CLcm  =  0.29459889225936597;    
      Real IAzz  =  0.17680943965733698;    
      Real IBzz  =  0.55702248757755590;    
      Real ICzz  =  1.7664179506435480;     
      Real mA    =  2.7272699999999999;     
      Real mB    =  8.3802599999999998;     
      Real mC    =  30.470580000000005;     


      // Query the matter subsystem for the time in the simulation
      const Real time = state.getTime();

       // set desired position, velocity, and acceleration*pow(t,5)
      
      Vec3 qAinfo;
      Vec3 qBinfo;
      Vec3 qCinfo;

      if ( time <= 10.0 )   // if statement #2
      {
         Real qADes = -45.0*3.14/180.0;
         Real qBDes = 90.0*3.14/180.0;
         Real qCDes = -45.0*3.14/180.0;
         qAinfo = calculateDesiredAngleAndDerivativeFromData( time, -10.0*3.14/180.0, qADes, 0.0, 10.0 );
         qBinfo = calculateDesiredAngleAndDerivativeFromData( time, 10.0*3.14/180.0, qBDes, 0.0, 10.0 );
         qCinfo = calculateDesiredAngleAndDerivativeFromData( time, -10.0*3.14/180.0, qCDes, 0.0, 10.0 );
         

      }
      else   // paired with if statement #2
      {
         Real qADes = -10.0*3.14/180.0;
         Real qBDes = 20.0*3.14/180.0;
         Real qCDes = -10.0*3.14/180.0;
         qAinfo = calculateDesiredAngleAndDerivativeFromData( time-10.0, -45.0*3.14/180.0, qADes, 0.0, 10.0 );
         qBinfo = calculateDesiredAngleAndDerivativeFromData( time-10.0, 90.0*3.14/180.0, qBDes, 0.0, 10.0 );
         qCinfo = calculateDesiredAngleAndDerivativeFromData( time-10.0, -45.0*3.14/180.0, qCDes, 0.0, 10.0 );
      }

      Real qADes = qAinfo[0];
		Real qBDes = qBinfo[0];
		Real qCDes = qCinfo[0];
      
		Real qApDes  = qAinfo[1];
		Real qBpDes  = qBinfo[1];
		Real qCpDes  = qCinfo[1];

		Real qAppDes  = qAinfo[2];
		Real qBppDes  = qBinfo[2];
		Real qCppDes  = qCinfo[2];

      // Set control constants
      Real kD = 2*omega*zeta;
      Real kP = omega*omega;

      // Determine desired acceleration plus error signal from velocity and position terms
		Real qApp = qAppDes + kD*(qApDes - qAp)+ kP*(qADes - qA);
		Real qBpp = qBppDes + kD*(qBpDes - qBp)+ kP*(qBDes - qB);
		Real qCpp = qCppDes + kD*(qCpDes - qCp)+ kP*(qCDes - qC);

      // Evaluate constants
      double z[30];
      z[1] = sin(qB);
      z[2] = sin(qB+qC);
      z[3] = sin(qC);
      z[4] = cos(qC);
      z[5] = cos(qB+qC);
      z[6] = cos(qB);
      z[7] = sin(qA);
      z[8] = sin(qA+qB);
      z[9] = sin(qA+qB+qC);
      z[10] = AL*BLcm*mB;
      z[11] = AL*CLcm*mC;
      z[12] = BL*CLcm*mC;
      z[13] = ALcm*mA;
      z[14] = AL*BL;
      z[15] = AL*CLcm;
      z[16] = BL*CLcm;
      z[17] = CLcm*mC;
      z[18] = IBzz + ICzz;
      z[19] = pow(BL,2);
      z[20] = IAzz + IBzz + ICzz + mB*pow(AL,2);
      z[21] = pow(AL,2) + pow(BL,2);
      z[22] = z[10]*z[1]*pow(qAp,2) + z[11]*z[2]*pow(qAp,2) + z[12]*z[3]*pow((qAp+
              qBp),2) + mC*(z[14]*z[1]*pow(qAp,2)-z[14]*z[1]*pow((qAp+qBp),2)-z[15]*z[2]*
              pow((qAp+qBp+qCp),2)-z[16]*z[3]*pow((qAp+qBp+qCp),2)) + (ICzz+z[17]*(AL*
              z[5]+BL*z[4]))*qCpp + (z[18]+z[10]*z[6]+2*z[12]*z[4]+mC*(z[19]+z[14]*z[6]+
              z[15]*z[5]))*qBpp + (z[20]+2*z[10]*z[6]+2*z[11]*z[5]+2*z[12]*z[4]+mC*(z[21]+
              2*z[14]*z[6]))*qApp - 9.810000000000001*z[13]*z[7] - 9.810000000000001*mB*(
              AL*z[7]+BLcm*z[8]) - 9.810000000000001*mC*(AL*z[7]+BL*z[8]+CLcm*z[9]) - 
              z[10]*z[1]*pow((qAp+qBp),2);
      z[23] = BL*mC;
      z[24] = BLcm*mB;
      z[25] = IBzz + ICzz + mC*pow(BL,2);
      z[26] = z[10]*z[1]*pow(qAp,2) + z[11]*z[2]*pow(qAp,2) + z[12]*z[3]*pow((qAp+
              qBp),2) + z[23]*(AL*z[1]*pow(qAp,2)-CLcm*z[3]*pow((qAp+qBp+qCp),2)) + (ICzz+
              z[12]*z[4])*qCpp + (z[25]+2*z[12]*z[4])*qBpp + (z[18]+z[10]*z[6]+z[11]*z[5]+
              2*z[12]*z[4]+z[23]*(BL+AL*z[6]))*qApp - 9.810000000000001*z[24]*z[8] - 
              9.810000000000001*mC*(BL*z[8]+CLcm*z[9]);
      z[27] = z[17]*(AL*z[2]*pow(qAp,2)+BL*z[3]*pow((qAp+qBp),2)) + ICzz*qCpp + (
              ICzz+z[12]*z[4])*qBpp + (ICzz+z[11]*z[5]+z[12]*z[4])*qApp - 9.810000000000001*
              z[17]*z[9];

      // Define joint torques to be applied
      double  TAnkle = z[22];
      double  TKnee  = z[26];
      double  THip   = z[27];

      // Get the proper memory location to increment the force and/or torque.
      // bodiesForces is a Vec6 whose elements are two Vec3.
      // The elements of the first  Vec3 are Tx, Ty, Tz  (expressed in the ground's "x,y,z").
      // The elements of the second Vec3 are Fx, Fy, Fz  (expressed in the ground's "x,y,z").
      SpatialVec& shankBodiesForces = bodyForces[ myShankId ];
      Vec3& shankTorqueSum = shankBodiesForces[0];
      Vec3& shankForceSum  = shankBodiesForces[1];
      shankTorqueSum[2] += TAnkle;       // Increment torque in the shank's z-direction.
      shankTorqueSum[2] -= TKnee;        // Equal & opposite reaction torque

	   SpatialVec& thighBodiesForces = bodyForces[ myThighId ];
      Vec3& thighTorqueSum = thighBodiesForces[0];
      Vec3& thighForceSum  = thighBodiesForces[1];
	   thighTorqueSum[2] += TKnee; 
      thighTorqueSum[2] -= THip;         // Equal & opposite reaction torque

	   SpatialVec& HATBodiesForces = bodyForces[ myHATId ];
      Vec3& HATTorqueSum = HATBodiesForces[0];
      Vec3& HATForceSum  = HATBodiesForces[1];
	   HATTorqueSum[2] += THip; 

      // Use cout to display torques to a txt file specified in Project Properties.
      // Project properties: Configuration properties: Debugging: Command Arguments: Type in ">filename.txt"
      cout<< time;
      cout<<"\t";
      cout<< TAnkle;
      cout<<"\t";
      cout<< TKnee;
      cout<<"\t";
      cout<< THip;
      cout<<"\n";
   }   
Vec3 calculateDesiredAngleAndDerivativeFromData( Real t, Real x0, Real xtf, Real xptf, Real tf ) const
{
   Real c0 = x0;
   Real  c1 = 0;
   Real  c2 = (3*xtf-3*x0-tf*xptf)/(pow(tf,2));
   Real c3 = -(2*xtf-2*x0-tf*xptf)/(pow(tf,3));

   Real x = c0 +c1*t+c2*pow(t,2)+c3*pow(t,3);
   Real xp = c1 + 2*c2*t + 3*c3*pow(t,2);
   Real xpp = 2*c2 +6*c3*t;

   return Vec3(x,xp,xpp);

}

private:
    BodyId  myShankId;
	 BodyId	myThighId;
	 BodyId	myHATId;
    Real    zeta;
    Real    omega;

}; 
//-----------------------------------------------------------------------------
#endif /* __USERFORCEVCONTROLHORIZONTAL_H__ */
//-----------------------------------------------------------------------------
